from pathlib import Path

from docling.datamodel.base_models import InputFormat
from docling.document_converter import DocumentConverter, FormatOption
from spacy.language import Language
from spacy.tokens import Doc, Span, SpanGroup

from .types import Attrs, DocLayout, PageLayout, SpanLayout


class spaCyLayout:
    def __init__(
        self,
        nlp: Language,
        separator: str | None = "\n\n",
        attrs: dict[str, str] = {},
        docling_options: dict[InputFormat, FormatOption] | None = None,
    ) -> None:
        """Initialize the layout parser and Docling converter."""
        self.nlp = nlp
        self.sep = separator
        self.attrs = Attrs(
            doc_layout=attrs.get("doc_layout", "layout"),
            doc_pages=attrs.get("doc_pages", "pages"),
            span_layout=attrs.get("span_layout", "layout"),
            span_group=attrs.get("span_group", "layout"),
        )
        self.converter = DocumentConverter(format_options=docling_options)
        # Set spaCy extension attributes for custom data
        Doc.set_extension(self.attrs.doc_layout, default=None, force=True)
        Doc.set_extension(self.attrs.doc_pages, getter=self.get_pages, force=True)
        Span.set_extension(self.attrs.span_layout, default=None, force=True)

    def __call__(self, path: str | Path) -> Doc:
        """Call parser on a path to create a spaCy Doc object."""
        result = self.converter.convert(path)
        inputs = []
        for item in result.document.texts:
            if item.prov:
                prov = item.prov[0]
                bounding_box = SpanLayout(
                    x=prov.bbox.l,
                    y=prov.bbox.t,
                    width=prov.bbox.r - prov.bbox.l,
                    height=prov.bbox.b - prov.bbox.t,
                    page_no=prov.page_no,
                )
            else:
                bounding_box = None
            inputs.append((item.text, item.label, bounding_box))
        doc = self._texts_to_doc(inputs)
        pages = [
            PageLayout(
                page_no=i + 1,
                width=page.size.width if page.size else 0,
                height=page.size.height if page.size else 0,
            )
            for i, page in enumerate(result.pages)
        ]
        doc._.set(self.attrs.doc_layout, DocLayout(pages=pages))
        return doc

    def _texts_to_doc(self, inputs: list[tuple[str, str, SpanLayout]]) -> Doc:
        """Convert Docling structure to spaCy Doc."""
        words = []
        spaces = []
        span_data = []
        token_idx = 0
        for item_text, label, layout in inputs:
            if item_text == "":
                continue
            # Tokenize the span because we can't rely on the document parsing to
            # give us items that are not split across token boundaries
            with self.nlp.select_pipes(disable=self.nlp.pipe_names):
                span_doc = self.nlp(item_text)
            words += [token.text for token in span_doc]
            spaces += [bool(token.whitespace_) for token in span_doc]
            # Add separator token and don't include it in the layout span
            if self.sep:
                words.append(self.sep)
                spaces[-1] = False
                spaces.append(False)
            end = token_idx + len(span_doc)
            span_data.append((token_idx, end, label, layout))
            token_idx += len(span_doc) + (1 if self.sep else 0)
        doc = Doc(self.nlp.vocab, words=words, spaces=spaces)
        spans = []
        for start, end, label, layout in span_data:
            span = Span(doc, start=start, end=end, label=label)
            span._.set(self.attrs.span_layout, layout)
            spans.append(span)
        doc.spans[self.attrs.span_group] = SpanGroup(
            doc, name=self.attrs.span_group, spans=spans
        )
        return doc

    def get_pages(self, doc: Doc) -> list[tuple[PageLayout, list[Span]]]:
        """Get all pages and their layout spans."""
        layout = doc._.get(self.attrs.doc_layout)
        pages = {page.page_no: page for page in layout.pages}
        page_spans = {page.page_no: [] for page in layout.pages}
        for span in doc.spans[self.attrs.span_group]:
            span_layout = span._.get(self.attrs.span_layout)
            page_spans[span_layout.page_no].append(span)
        return [(pages[i], page_spans[i]) for i in page_spans]



from dataclasses import dataclass


@dataclass
class Attrs:
    """Custom atributes used to extend spaCy"""

    doc_layout: str
    doc_pages: str
    span_layout: str
    span_group: str


@dataclass
class PageLayout:
    page_no: int
    width: float
    height: float


@dataclass
class DocLayout:
    """Document layout features added to Doc object"""

    pages: list[PageLayout]


@dataclass
class SpanLayout:
    """Text span layout features added to Span object"""

    x: float
    y: float
    width: float
    height: float
    page_no: int



from pathlib import Path

import pytest
import spacy
from docling_core.types.doc.labels import DocItemLabel

from spacy_layout import spaCyLayout
from spacy_layout.types import DocLayout, SpanLayout

PDF_STARCRAFT = Path(__file__).parent / "data" / "starcraft.pdf"
PDF_SIMPLE = Path(__file__).parent / "data" / "simple.pdf"
DOCX_SIMPLE = Path(__file__).parent / "data" / "simple.docx"


@pytest.fixture
def nlp():
    return spacy.blank("en")


@pytest.fixture
def span_labels():
    return [label.value for label in DocItemLabel]


@pytest.mark.parametrize("path", [PDF_STARCRAFT, PDF_SIMPLE])
def test_general(path, nlp, span_labels):
    layout = spaCyLayout(nlp)
    doc = layout(path)
    assert isinstance(doc._.get(layout.attrs.doc_layout), DocLayout)
    for span in doc.spans[layout.attrs.span_group]:
        assert span.text
        assert span.label_ in span_labels
        assert isinstance(span._.get(layout.attrs.span_layout), SpanLayout)


@pytest.mark.parametrize("path", [PDF_SIMPLE, DOCX_SIMPLE])
@pytest.mark.parametrize("separator", ["\n\n", ""])
def test_simple(path, separator, nlp):
    layout = spaCyLayout(nlp, separator=separator)
    doc = layout(path)
    assert len(doc.spans[layout.attrs.span_group]) == 4
    assert doc.text.startswith(f"Lorem ipsum dolor sit amet{separator}")
    assert doc.spans[layout.attrs.span_group][0].text == "Lorem ipsum dolor sit amet"
